﻿Option Strict On
Option Explicit On

Imports System.Runtime.InteropServices

Public Class OCR
    Implements IDisposable

#Region "API"
    <DllImport("Combase.dll", EntryPoint:="WindowsCreateString")>
    <PreserveSig> Private Shared Function WindowsCreateString(<[In], MarshalAs(UnmanagedType.LPWStr)> sourceString As String,
                                                              <[In]> length As Integer,
                                                              <Out> ByRef hString As IntPtr) As Integer
    End Function

    <DllImport("Combase.dll", EntryPoint:="WindowsGetStringRawBuffer")>
    <PreserveSig> Private Shared Function WindowsGetStringRawBuffer(<[In]> hstring As IntPtr,
                                                                    <Out> ByRef length As Integer) As IntPtr
    End Function

    <DllImport("Combase.dll", EntryPoint:="WindowsDeleteString")>
    <PreserveSig> Private Shared Function WindowsDeleteString(<[In]> hString As IntPtr) As Integer
    End Function

    <DllImport("Combase.dll", EntryPoint:="RoGetActivationFactory")>
    <PreserveSig> Private Shared Function RoGetActivationFactory(<[In]> activatableClassId As IntPtr,
                                                                 <[In], MarshalAs(UnmanagedType.LPStruct)> riid As Guid,
                                                                 <Out> ByRef factory As IntPtr) As Integer

    End Function

    <DllImport("Shlwapi.dll", EntryPoint:="SHCreateMemStream")>
    Private Shared Function SHCreateMemStream(<[In]> pInit As IntPtr,
                                              <[In]> cbInit As UInteger) As IntPtr
    End Function


    <DllImport("ShCore.dll", EntryPoint:="CreateRandomAccessStreamOnFile")>
    <PreserveSig> Private Shared Function CreateRandomAccessStreamOnFile(<[In], MarshalAs(UnmanagedType.LPWStr)> filePath As String,
                                                                         <[In]> accessMode As FileAccessMode,
                                                                         <[In], MarshalAs(UnmanagedType.LPStruct)> riid As Guid,
                                                                         <Out> ByRef ppv As IntPtr) As Integer
    End Function

    <DllImport("ShCore.dll", EntryPoint:="CreateRandomAccessStreamOverStream")>
    <PreserveSig> Private Shared Function CreateRandomAccessStreamOverStream(<[In]> pStream As IntPtr,
                                                                             <[In]> options As BSOS_OPTIONS,
                                                                             <[In], MarshalAs(UnmanagedType.LPStruct)> riid As Guid,
                                                                             <Out> ByRef ppv As IntPtr) As Integer
    End Function

    <DllImport("GdiPlus.dll", EntryPoint:="GdipCreateBitmapFromHBITMAP")>
    Private Shared Function GdipCreateBitmapFromHBITMAP(<[In]> hbm As IntPtr,
                                                        <[In]> hpal As IntPtr,
                                                        <Out> ByRef pBitmap As IntPtr) As Integer
    End Function

    <DllImport("GdiPlus.dll", EntryPoint:="GdipDisposeImage")>
    Private Shared Function GdipDisposeImage(<[In]> pImage As IntPtr) As Integer
    End Function

    <DllImport("GdiPlus.dll", EntryPoint:="GdipSaveImageToStream")>
    Private Shared Function GdipSaveImageToStream(<[In]> pImage As IntPtr,
                                                  <[In]> pStream As IntPtr,
                                                  <[In]> ByRef ClsidEncoder As Guid,
                                                  <[In]> EncoderParams As IntPtr) As Integer
    End Function

    <DllImport("Gdi32.dll", EntryPoint:="DeleteObject")>
    Private Shared Function DeleteObject(<[In]> hObject As IntPtr) As <MarshalAs(UnmanagedType.Bool)> Boolean
    End Function
#End Region

#Region "Enum"
    Private Enum VTable_Interfaces As Integer

        'IReadOnlyList
        IReadOnlyList_GetAt = 6
        IReadOnlyList_GetSize = 7

        ' ILanguage
        ILanguage_GetLanguageTag = 6
        ILanguage_GetDisplayName = 7
        ILanguage_GetNativeName = 8
        ILanguage_GetScript = 9

        ' IOcrEngineStatics
        IOcrEngineStatics_GetMaxImageDimension = 6
        IOcrEngineStatics_GetAvailableRecognizerLanguages = 7
        IOcrEngineStatics_IsLanguageSupported = 8
        IOcrEngineStatics_TryCreateFromLanguage = 9
        IOcrEngineStatics_TryCreateFromUserProfileLanguages = 10

        ' IOcrEngine
        IOcrEngine_RecognizeAsync = 6
        IOcrEngine_GetRecognizerLanguage = 7

        ' IOcrResult
        IOcrResult_GetLines = 6
        IOcrResult_GetTextAngle = 7
        IOcrResult_GetText = 8

        ' IOcrLine
        IOcrLine_GetWords = 6
        IOcrLine_GetText = 7

        ' IOcrWord
        IOcrWord_GetBoundingRect = 6
        IOcrWord_GetText = 7

        ' IBitmapDecoderStatics
        IBitmapDecoderStatics_CreateAsync = 14

        ' IBitmapFrame
        IBitmapFrame_GetPixelWidth = 12
        IBitmapFrame_GetPixelHeight = 13

        ' IBitmapFrameWithSoftwareBitmap
        IBitmapFrameWithSoftwareBitmap_GetSoftwareBitmapAsync = 6

        ' IAsyncInfo
        IAsyncInfo_GetStatus = 7
        IAsyncInfo_Close = 10

        ' IAsyncInterface
        IAsyncInterface_GetResults = 8

        ' IClosable
        IClosable_Close = 6
    End Enum

    Private Enum FileAccessMode As Integer
        Read = 0
        ReadWrite = 1
    End Enum

    Private Enum BSOS_OPTIONS As Integer
        BSOS_DEFAULT = 0
        BSOS_PREFERDESTINATIONSTREAM = 1
    End Enum

    Private Enum AsyncStatus As Integer
        Started = 0
        Completed = 1
        Canceled = 2
        [Error] = 3
    End Enum
#End Region

#Region "Const"
    Private Const S_OK As Integer = 0
    Private Const GDIP_OK As Integer = 0

    Private Const WindowsMediaOcrOcrEngine As String = "Windows.Media.Ocr.OcrEngine"
    Private Const WindowsGraphicsImagingBitmapDecoder As String = "Windows.Graphics.Imaging.BitmapDecoder"

    Private Const IID_IClosable As String = "30d5a829-7fa4-4026-83bb-d75bae4ea99e"
    Private Const IID_IAsyncInfo As String = "00000036-0000-0000-c000-000000000046"
    Private Const IID_IBitmapFrame As String = "72a49a1c-8081-438d-91bc-94ecfc8185c6"
    Private Const IID_IOcrEngineStatics As String = "5bffa85a-3384-3540-9940-699120d428a8"
    Private Const IID_IRandomAccessStream As String = "905a0fe1-bc53-11df-8c49-001e4fc686da"
    Private Const IID_IBitmapDecoderStatics As String = "438ccb26-bcef-4e95-bad6-23a822e58d01"
    Private Const IID_IBitmapFrameWithSoftwareBitmap As String = "fe287c9a-420c-4963-87ad-691436e08383"

    Private Const CLSID_GDIPPngEncoder As String = "557cf406-1a04-11d3-9a73-0000f81ef32e"
#End Region

#Region "Class"
    Sub New()
        If GetActivationFactory(WindowsMediaOcrOcrEngine,
                                IID_IOcrEngineStatics,
                                m_pIOcrEngineStatics) Then
            If GetActivationFactory(WindowsGraphicsImagingBitmapDecoder,
                                    IID_IBitmapDecoderStatics,
                                    m_pIBitmapDecoderStatics) Then
                m_bolIsOcrInitialized = True
                OcrEngineStatics_GetMaxImageDimension(m_pIOcrEngineStatics, m_intMaxImageDimension)
                GetAvailableOcrLanguages()
            End If
        End If
    End Sub
#End Region

#Region "Variablen"
    Private m_bolDisposedValue As Boolean
    Private ReadOnly m_bolIsOcrInitialized As Boolean
    Private ReadOnly m_intMaxImageDimension As Integer
    Private m_tOcrResults As OcrResults
    Private m_listAvailableOcrLanguages As List(Of OcrLanguage)
    Private m_pIOcrEngine As IntPtr
    Private m_pIOcrEngineStatics As IntPtr
    Private m_pIBitmapDecoderStatics As IntPtr
#End Region

#Region "Structure"
    Private Structure OcrLanguage
        Dim LanguageTag As String
        Dim DisplayName As String
        Dim NativeName As String
        Dim Script As String
    End Structure

    Private Structure OcrWord
        Dim TextLineWord As String
        Dim TextLineWordRectF As RectangleF
    End Structure

    Private Structure OcrLine
        Dim TextLine As String
        Dim TextLineWords As List(Of OcrWord)
    End Structure

    Private Structure OcrResults
        Dim TextOcrLanguage As OcrLanguage
        Dim TextAngle As Double
        Dim TextComplete As String
        Dim TextLines As List(Of OcrLine)
    End Structure

#End Region

#Region "Properties"
    Public ReadOnly Property GetIsOcrInitialized As Boolean
        Get
            Return m_bolIsOcrInitialized
        End Get
    End Property

    Public ReadOnly Property GetMaxOcrImageDimension As Integer
        Get
            Return m_intMaxImageDimension
        End Get
    End Property

    Public ReadOnly Property GetAvailableOcrLanguagesCount As Integer
        Get
            Return m_listAvailableOcrLanguages.Count
        End Get
    End Property

    Public ReadOnly Property GetAvailableOcrLanguageTag(Index As Integer) As String
        Get
            If Index >= 0 AndAlso Index < Me.GetAvailableOcrLanguagesCount Then
                Return m_listAvailableOcrLanguages(Index).LanguageTag
            Else
                Return String.Empty
            End If
        End Get
    End Property

    Public ReadOnly Property GetAvailableOcrLanguageDisplayName(Index As Integer) As String
        Get
            If Index >= 0 AndAlso Index < Me.GetAvailableOcrLanguagesCount Then
                Return m_listAvailableOcrLanguages(Index).DisplayName
            Else
                Return String.Empty
            End If
        End Get
    End Property

    Public ReadOnly Property GetAvailableOcrLanguageNativeName(Index As Integer) As String
        Get
            If Index >= 0 AndAlso Index < Me.GetAvailableOcrLanguagesCount Then
                Return m_listAvailableOcrLanguages(Index).NativeName
            Else
                Return String.Empty
            End If
        End Get
    End Property

    Public ReadOnly Property GetAvailableOcrLanguageScript(Index As Integer) As String
        Get
            If Index >= 0 AndAlso Index < Me.GetAvailableOcrLanguagesCount Then
                Return m_listAvailableOcrLanguages(Index).Script
            Else
                Return String.Empty
            End If
        End Get
    End Property

    Public ReadOnly Property InitOcrEngineByLanguageIndex(Index As Integer) As Boolean
        Get
            If Index >= 0 AndAlso Index < Me.GetAvailableOcrLanguagesCount Then
                Return InitOcrEngineByIndex(Index)
            Else
                Return False
            End If
        End Get
    End Property

    Public ReadOnly Property GetUsedOcrResultLanguageTag() As String
        Get
            Return m_tOcrResults.TextOcrLanguage.LanguageTag
        End Get
    End Property

    Public ReadOnly Property GetUsedOcrResultLanguageDisplayName() As String
        Get
            Return m_tOcrResults.TextOcrLanguage.DisplayName
        End Get
    End Property

    Public ReadOnly Property GetUsedOcrResultLanguageNativeName() As String
        Get
            Return m_tOcrResults.TextOcrLanguage.NativeName
        End Get
    End Property

    Public ReadOnly Property GetUsedOcrResultLanguageScript() As String
        Get
            Return m_tOcrResults.TextOcrLanguage.Script
        End Get
    End Property

    Public ReadOnly Property GetOcrResultTextAngle() As Double
        Get
            Return m_tOcrResults.TextAngle
        End Get
    End Property

    Public ReadOnly Property GetOcrResultTextComplete() As String
        Get
            Return m_tOcrResults.TextComplete
        End Get
    End Property

    Public ReadOnly Property GetOcrResultTextCompleteWithLineBreak() As String
        Get
            Dim intCount As Integer = Me.GetOcrResultTextLineCount
            Dim strRet As String = String.Empty
            If intCount > 0 Then
                For intIndex As Integer = 0 To intCount - 1
                    strRet &= Me.GetOcrResultTextLineByIndex(intIndex) &
                        If(intIndex < intCount - 1, Environment.NewLine, String.Empty)
                Next
            End If
            Return strRet
        End Get
    End Property

    Public ReadOnly Property GetOcrResultTextLineCount() As Integer
        Get
            Return m_tOcrResults.TextLines.Count
        End Get
    End Property

    Public ReadOnly Property GetOcrResultTextLineByIndex(LineIndex As Integer) As String
        Get
            If LineIndex >= 0 AndAlso LineIndex < Me.GetOcrResultTextLineCount Then
                Return m_tOcrResults.TextLines(LineIndex).TextLine
            Else
                Return String.Empty
            End If
        End Get
    End Property

    Public ReadOnly Property GetOcrResultTextLineByIndexWordsCount(LineIndex As Integer) As Integer
        Get
            If LineIndex >= 0 AndAlso LineIndex < Me.GetOcrResultTextLineCount Then
                Return m_tOcrResults.TextLines(LineIndex).TextLineWords.Count
            Else
                Return 0
            End If
        End Get
    End Property

    Public ReadOnly Property GetOcrResultTextLineByIndexWordByIndex(LineIndex As Integer,
                                                                    WordIndex As Integer) As String
        Get
            If LineIndex >= 0 AndAlso LineIndex < Me.GetOcrResultTextLineCount AndAlso
                WordIndex >= 0 AndAlso WordIndex < Me.GetOcrResultTextLineByIndexWordsCount(LineIndex) Then
                Return m_tOcrResults.TextLines(LineIndex).TextLineWords(WordIndex).TextLineWord
            Else
                Return String.Empty
            End If
        End Get
    End Property

    Public ReadOnly Property GetOcrResultTextLineByIndexWordRectByIndex(LineIndex As Integer,
                                                                        WordIndex As Integer) As RectangleF
        Get
            If LineIndex >= 0 AndAlso LineIndex < Me.GetOcrResultTextLineCount AndAlso
                WordIndex >= 0 AndAlso WordIndex < Me.GetOcrResultTextLineByIndexWordsCount(LineIndex) Then
                Return m_tOcrResults.TextLines(LineIndex).TextLineWords(WordIndex).TextLineWordRectF
            Else
                Return New RectangleF()
            End If
        End Get
    End Property

#End Region

#Region "Public Functions"
    Public Function GetOcrResultsFromImageFile(ImageFile As String) As Boolean
        Dim bolRet As Boolean = False
        m_tOcrResults = New OcrResults
        If m_bolIsOcrInitialized Then
            If m_pIOcrEngine <> IntPtr.Zero AndAlso Not String.IsNullOrEmpty(ImageFile) Then
                Dim pIRandomAccessStream As IntPtr = IntPtr.Zero
                If CreateRandomAccessStreamOnFile(ImageFile, FileAccessMode.Read,
                                                  New Guid(IID_IRandomAccessStream),
                                                  pIRandomAccessStream) = S_OK Then
                    bolRet = GetOcrResults(pIRandomAccessStream)
                    CloseAndRelease(pIRandomAccessStream)
                End If
            End If
        End If
        Return bolRet
    End Function

    Public Function GetOcrResultsFromBitmap(bitmap As Bitmap) As Boolean
        Dim bolRet As Boolean = False
        m_tOcrResults = New OcrResults
        If m_bolIsOcrInitialized Then
            If m_pIOcrEngine <> IntPtr.Zero AndAlso bitmap IsNot Nothing Then
                Dim hBitmap As IntPtr = bitmap.GetHbitmap
                If hBitmap <> IntPtr.Zero Then
                    Dim pBitmap As IntPtr = IntPtr.Zero
                    If GdipCreateBitmapFromHBITMAP(hBitmap, IntPtr.Zero, pBitmap) = GDIP_OK Then
                        Dim pIStream As IntPtr = SHCreateMemStream(IntPtr.Zero, 0)
                        If pIStream <> IntPtr.Zero Then
                            If GdipSaveImageToStream(pBitmap, pIStream,
                                                 New Guid(CLSID_GDIPPngEncoder),
                                                 IntPtr.Zero) = GDIP_OK Then
                                Dim pIRandomAccessStream As IntPtr = IntPtr.Zero
                                If CreateRandomAccessStreamOverStream(pIStream, BSOS_OPTIONS.BSOS_DEFAULT,
                                                                  New Guid(IID_IRandomAccessStream),
                                                                  pIRandomAccessStream) = S_OK Then
                                    bolRet = GetOcrResults(pIRandomAccessStream)
                                    CloseAndRelease(pIRandomAccessStream)
                                End If
                            End If
                            Release(pIStream)
                        End If
                        GdipDisposeImage(pBitmap)
                    End If
                    DeleteObject(hBitmap)
                End If
            End If
        End If
        Return bolRet
    End Function
#End Region

#Region "Private Functions"
    Private Function GetOcrResults(pIRandomAccessStream As IntPtr) As Boolean
        Dim bolRet As Boolean = False
        Dim pISoftwareBitmap As IntPtr = IntPtr.Zero
        If GetSoftwareBitmapFromRandomAccessStream(pIRandomAccessStream, pISoftwareBitmap) Then
            Dim pIOcrResult As IntPtr = IntPtr.Zero
            If OcrEngine_RecognizeAsync(m_pIOcrEngine, pISoftwareBitmap, pIOcrResult) Then
                If WaitForAsync(pIOcrResult) Then
                    GetUsedOcrLanguage()
                    OcrResult_GetTextAngle(pIOcrResult, m_tOcrResults.TextAngle)
                    Dim hString As IntPtr = IntPtr.Zero
                    If OcrResult_GetText(pIOcrResult, hString) Then
                        m_tOcrResults.TextComplete = Marshal.PtrToStringUni(WindowsGetStringRawBuffer(hString, 0))
                    End If
                    bolRet = GetTextLinesFromOcrResult(pIOcrResult)
                End If
                Release(pIOcrResult)
            End If
            CloseAndRelease(pISoftwareBitmap)
        End If
        Return bolRet
    End Function

    Private Function GetTextLinesFromOcrResult(pIOcrResult As IntPtr) As Boolean
        Dim bolRet As Boolean = False
        Dim pIReadOnlyList As IntPtr = IntPtr.Zero
        If OcrResult_GetLines(pIOcrResult, pIReadOnlyList) Then
            Dim intCount As Integer
            If ReadOnlyList_GetSize(pIReadOnlyList, intCount) Then
                If intCount > 0 Then
                    m_tOcrResults.TextLines = New List(Of OcrLine)
                    For intIndex As Integer = 0 To intCount - 1
                        Dim tOcrLine As New OcrLine
                        Dim pIOcrLine As IntPtr = IntPtr.Zero
                        If ReadOnlyList_GetAt(pIReadOnlyList, intIndex, pIOcrLine) Then
                            Dim hString As IntPtr = IntPtr.Zero
                            If OcrLine_GetText(pIOcrLine, hString) Then
                                tOcrLine.TextLine = Marshal.PtrToStringUni(WindowsGetStringRawBuffer(hString, 0))
                                GetWordsAndPositionromTextLine(pIOcrLine, tOcrLine)
                            End If
                            Release(pIOcrLine)
                        End If
                        m_tOcrResults.TextLines.Add(tOcrLine)
                    Next
                    bolRet = True
                End If
            End If
            Release(pIReadOnlyList)
        End If
        Return bolRet
    End Function

    Private Sub GetWordsAndPositionromTextLine(pIOcrLine As IntPtr,
                                               ByRef tOcrLine As OcrLine)
        Dim pIReadOnlyList As IntPtr = IntPtr.Zero
        If OcrLine_GetWords(pIOcrLine, pIReadOnlyList) Then
            Dim intCount As Integer
            If ReadOnlyList_GetSize(pIReadOnlyList, intCount) Then
                If intCount > 0 Then
                    tOcrLine.TextLineWords = New List(Of OcrWord)
                    For intIndex As Integer = 0 To intCount - 1
                        Dim pIOcrWord As IntPtr = IntPtr.Zero
                        Dim tOcrWord As New OcrWord
                        If ReadOnlyList_GetAt(pIReadOnlyList, intIndex, pIOcrWord) Then
                            Dim tRectF As New RectangleF
                            If OcrWord_GetBoundingRect(pIOcrWord, tRectF) Then
                                tOcrWord.TextLineWordRectF = tRectF
                                Dim hString As IntPtr = IntPtr.Zero
                                If OcrWord_GetText(pIOcrWord, hString) Then
                                    tOcrWord.TextLineWord = Marshal.PtrToStringUni(WindowsGetStringRawBuffer(hString, 0))
                                End If
                            End If
                            Release(pIOcrWord)
                        End If
                        tOcrLine.TextLineWords.Add(tOcrWord)
                    Next
                End If
            End If
            Release(pIReadOnlyList)
        End If
    End Sub

    Private Function GetSoftwareBitmapFromRandomAccessStream(pIRandomAccessStream As IntPtr,
                                                             ByRef pISoftwareBitmap As IntPtr) As Boolean
        Dim bolRet As Boolean = False
        If m_bolIsOcrInitialized Then
            Dim pIBitmapDecoder As IntPtr = IntPtr.Zero
            If BitmapDecoderStatics_CreateAsync(m_pIBitmapDecoderStatics,
                                                pIRandomAccessStream,
                                                pIBitmapDecoder) Then
                If WaitForAsync(pIBitmapDecoder) Then
                    Dim pIBitmapFrame As IntPtr = IntPtr.Zero
                    If QueryInterface(pIBitmapDecoder,
                                      IID_IBitmapFrame,
                                      pIBitmapFrame) Then
                        Dim intFrameWidth As Integer
                        Dim intFrameHeight As Integer
                        BitmapFrame_GetPixelWidth(pIBitmapFrame, intFrameWidth)
                        BitmapFrame_GetPixelHeight(pIBitmapFrame, intFrameHeight)
                        If intFrameWidth > m_intMaxImageDimension OrElse
                            intFrameHeight > m_intMaxImageDimension Then
                            MessageBox.Show("Image is to big. Maximum: " &
                                            m_intMaxImageDimension.ToString & "x" &
                                            m_intMaxImageDimension.ToString & " Pixel.", "OCR",
                                            MessageBoxButtons.OK, MessageBoxIcon.Error)
                        Else
                            Dim pIBitmapFrameWithSoftwareBitmap As IntPtr = IntPtr.Zero
                            If QueryInterface(pIBitmapDecoder,
                                              IID_IBitmapFrameWithSoftwareBitmap,
                                              pIBitmapFrameWithSoftwareBitmap) Then
                                If BitmapFrameWithSoftwareBitmap_GetSoftwareBitmapAsync(pIBitmapFrameWithSoftwareBitmap,
                                                                                        pISoftwareBitmap) Then
                                    If WaitForAsync(pISoftwareBitmap) Then bolRet = True
                                End If
                                Release(pIBitmapFrameWithSoftwareBitmap)
                            End If
                        End If
                        Release(pIBitmapFrame)
                    End If
                End If
                Release(pIBitmapDecoder)
            End If
        End If
        Return bolRet
    End Function

    Private Sub GetUsedOcrLanguage()
        If m_bolIsOcrInitialized AndAlso m_pIOcrEngine <> IntPtr.Zero Then
            Dim pILanguage As IntPtr = IntPtr.Zero
            If OcrEngine_GetRecognizerLanguage(m_pIOcrEngine, pILanguage) Then
                Dim hString As IntPtr = IntPtr.Zero
                If Language_GetLanguageTag(pILanguage, hString) Then
                    m_tOcrResults.TextOcrLanguage.LanguageTag =
                        Marshal.PtrToStringUni(WindowsGetStringRawBuffer(hString, 0))
                End If
                If Language_GetDisplayName(pILanguage, hString) Then
                    m_tOcrResults.TextOcrLanguage.DisplayName =
                        Marshal.PtrToStringUni(WindowsGetStringRawBuffer(hString, 0))
                End If
                If Language_GetNativeName(pILanguage, hString) Then
                    m_tOcrResults.TextOcrLanguage.NativeName =
                        Marshal.PtrToStringUni(WindowsGetStringRawBuffer(hString, 0))
                End If
                If Language_GetScript(pILanguage, hString) Then
                    m_tOcrResults.TextOcrLanguage.Script =
                        Marshal.PtrToStringUni(WindowsGetStringRawBuffer(hString, 0))
                End If
                Release(pILanguage)
            End If
        End If
    End Sub

    Private Function InitOcrEngineByIndex(Index As Integer) As Boolean
        Dim bolRet As Boolean = False
        If m_bolIsOcrInitialized Then
            If Index >= 0 AndAlso Index < Me.GetAvailableOcrLanguagesCount Then
                Release(m_pIOcrEngine)
                Dim pILanguage As IntPtr = IntPtr.Zero
                If GetLanguageByIndex(Index, pILanguage) Then
                    Dim bolIsLanguageSupported As Boolean = False
                    If OcrEngineStatics_IsLanguageSupported(m_pIOcrEngineStatics,
                                                            pILanguage,
                                                            bolIsLanguageSupported) Then
                        If bolIsLanguageSupported Then
                            If OcrEngineStatics_TryCreateFromLanguage(m_pIOcrEngineStatics,
                                                                      pILanguage,
                                                                      m_pIOcrEngine) Then bolRet = True
                        End If
                    End If
                    Release(pILanguage)
                End If
            End If
        End If
        Return bolRet
    End Function

    Private Function GetLanguageByIndex(Index As Integer,
                                        ByRef pILanguage As IntPtr) As Boolean
        Dim bolRet As Boolean = False
        If m_bolIsOcrInitialized Then
            If Index >= 0 AndAlso Index < Me.GetAvailableOcrLanguagesCount Then
                Dim pIReadOnlyList As IntPtr = IntPtr.Zero
                If OcrEngineStatics_GetAvailableRecognizerLanguages(m_pIOcrEngineStatics,
                                                                    pIReadOnlyList) Then
                    If ReadOnlyList_GetAt(pIReadOnlyList, Index,
                                          pILanguage) Then bolRet = True
                    Release(pIReadOnlyList)
                End If
            End If
        End If
        Return bolRet
    End Function

    Private Sub GetAvailableOcrLanguages()
        If m_bolIsOcrInitialized Then
            Dim pIReadOnlyList As IntPtr = IntPtr.Zero
            If OcrEngineStatics_GetAvailableRecognizerLanguages(m_pIOcrEngineStatics,
                                                                pIReadOnlyList) Then
                Dim intCount As Integer
                If ReadOnlyList_GetSize(pIReadOnlyList, intCount) Then
                    If intCount > 0 Then
                        m_listAvailableOcrLanguages = New List(Of OcrLanguage)
                        For intIndex As Integer = 0 To intCount - 1
                            Dim tOcrLanguage As New OcrLanguage
                            Dim pILanguage As IntPtr = IntPtr.Zero
                            If ReadOnlyList_GetAt(pIReadOnlyList, intIndex, pILanguage) Then
                                Dim hString As IntPtr = IntPtr.Zero
                                If Language_GetLanguageTag(pILanguage, hString) Then
                                    tOcrLanguage.LanguageTag =
                                        Marshal.PtrToStringUni(WindowsGetStringRawBuffer(hString, 0))
                                End If
                                If Language_GetDisplayName(pILanguage, hString) Then
                                    tOcrLanguage.DisplayName =
                                        Marshal.PtrToStringUni(WindowsGetStringRawBuffer(hString, 0))
                                End If
                                If Language_GetNativeName(pILanguage, hString) Then
                                    tOcrLanguage.NativeName =
                                        Marshal.PtrToStringUni(WindowsGetStringRawBuffer(hString, 0))
                                End If
                                If Language_GetScript(pILanguage, hString) Then
                                    tOcrLanguage.Script =
                                        Marshal.PtrToStringUni(WindowsGetStringRawBuffer(hString, 0))
                                End If
                                Release(pILanguage)
                            End If
                            m_listAvailableOcrLanguages.Add(tOcrLanguage)
                        Next
                    End If
                End If
                Release(pIReadOnlyList)
            End If
        End If
    End Sub

    Private Function GetActivationFactory(ClassName As String,
                                          IID As String,
                                          ByRef pFactory As IntPtr) As Boolean
        Dim bolRet As Boolean = False
        Dim hString As IntPtr = IntPtr.Zero
        If WindowsCreateString(ClassName, ClassName.Length, hString) = S_OK Then
            If hString <> IntPtr.Zero Then
                If RoGetActivationFactory(hString, New Guid(IID), pFactory) = S_OK Then bolRet = True
                WindowsDeleteString(hString)
            End If
        End If
        Return bolRet
    End Function

    Private Function WaitForAsync(ByRef pInterface As IntPtr) As Boolean
        Dim bolret As Boolean = False
        If pInterface <> IntPtr.Zero Then
            Dim pIAsyncInfo As IntPtr = IntPtr.Zero
            If QueryInterface(pInterface, IID_IAsyncInfo, pIAsyncInfo) Then
                Dim eStatus As AsyncStatus
                Do
                    AsyncInfo_GetStatus(pIAsyncInfo, eStatus)
                Loop While eStatus = AsyncStatus.Started
                If eStatus = AsyncStatus.Completed Then
                    bolret = True
                Else
                    AsyncInfo_Close(pIAsyncInfo)
                End If
                Release(pIAsyncInfo)
                If bolret Then
                    Dim pRet As IntPtr = IntPtr.Zero
                    If AsyncInterface_GetResults(pInterface, pRet) Then
                        Release(pInterface)
                        pInterface = pRet
                    End If
                End If
            End If
        End If
        Return bolret
    End Function

    Private Function GetVTablePtr(pInterface As IntPtr,
                                  intMethodNumber As VTable_Interfaces,
                                  ByRef pVTablePtr As IntPtr) As Boolean
        Dim bolRet As Boolean = False
        If pInterface <> IntPtr.Zero AndAlso intMethodNumber >= 0 Then
            Dim pVTablePtrs As IntPtr() = New IntPtr(0) {}
            Marshal.Copy(IntPtr.Add(Marshal.ReadIntPtr(pInterface),
                                    IntPtr.Size * intMethodNumber),
                                    pVTablePtrs, 0, 1)
            pVTablePtr = pVTablePtrs(0)
            bolRet = True
        End If
        Return bolRet
    End Function

    Private Sub Release(ByRef pInterface As IntPtr)
        If pInterface <> IntPtr.Zero Then
            Marshal.Release(pInterface)
            pInterface = IntPtr.Zero
        End If
    End Sub

    Private Sub CloseAndRelease(ByRef pInterface As IntPtr)
        If pInterface <> IntPtr.Zero Then
            Dim pIClosable As IntPtr = IntPtr.Zero
            If QueryInterface(pInterface, IID_IClosable, pIClosable) Then
                If Closable_Close(pIClosable) Then
                    Marshal.Release(pIClosable)
                    Marshal.Release(pInterface)
                    pInterface = IntPtr.Zero
                End If
            End If
        End If
    End Sub

    Private Function QueryInterface(pInterface As IntPtr,
                                    IID As String,
                                    ByRef pIOut As IntPtr) As Boolean
        Dim bolRet As Boolean = False
        If pInterface <> IntPtr.Zero Then
            If Marshal.QueryInterface(pInterface, New Guid(IID), pIOut) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function
#End Region

#Region "Interface Delegates"


    Private Delegate Function IReadOnlyList_GetAt(<[In]> this As IntPtr,
                                                  <[In]> index As Integer,
                                                  <Out> ByRef value As IntPtr) As Integer

    Private Delegate Function IReadOnlyList_GetSize(<[In]> this As IntPtr,
                                                    <Out> ByRef value As Integer) As Integer

    Private Delegate Function ILanguage_GetLanguageTag(<[In]> this As IntPtr,
                                                       <Out> ByRef value As IntPtr) As Integer

    Private Delegate Function ILanguage_GetDisplayName(<[In]> this As IntPtr,
                                                       <Out> ByRef value As IntPtr) As Integer

    Private Delegate Function ILanguage_GetNativeName(<[In]> this As IntPtr,
                                                      <Out> ByRef value As IntPtr) As Integer

    Private Delegate Function ILanguage_GetScript(<[In]> this As IntPtr,
                                                  <Out> ByRef value As IntPtr) As Integer

    Private Delegate Function IOcrEngineStatics_GetMaxImageDimension(<[In]> this As IntPtr,
                                                                     <Out> ByRef value As Integer) As Integer

    Private Delegate Function IOcrEngineStatics_GetAvailableRecognizerLanguages(<[In]> this As IntPtr,
                                                                                <Out> ByRef value As IntPtr) As Integer

    Private Delegate Function IOcrEngineStatics_IsLanguageSupported(<[In]> this As IntPtr,
                                                                    <[In]> language As IntPtr,
                                                                    <Out, MarshalAs(UnmanagedType.Bool)> ByRef value As Boolean) As Integer

    Private Delegate Function IOcrEngineStatics_TryCreateFromLanguage(<[In]> this As IntPtr,
                                                                      <[In]> language As IntPtr,
                                                                      <Out> ByRef result As IntPtr) As Integer

    Private Delegate Function IOcrEngine_RecognizeAsync(<[In]> this As IntPtr,
                                                        <[In]> bitmap As IntPtr,
                                                        <Out> ByRef result As IntPtr) As Integer

    Private Delegate Function IOcrEngine_GetRecognizerLanguage(<[In]> this As IntPtr,
                                                               <Out> ByRef value As IntPtr) As Integer

    Private Delegate Function IOcrResult_GetLines(<[In]> this As IntPtr,
                                                  <Out> ByRef value As IntPtr) As Integer

    Private Delegate Function IOcrResult_GetTextAngle(<[In]> this As IntPtr,
                                                      <Out> ByRef value As Double) As Integer

    Private Delegate Function IOcrResult_GetText(<[In]> this As IntPtr,
                                                 <Out> ByRef value As IntPtr) As Integer

    Private Delegate Function IOcrLine_GetWords(<[In]> this As IntPtr,
                                                 <Out> ByRef value As IntPtr) As Integer

    Private Delegate Function IOcrLine_GetText(<[In]> this As IntPtr,
                                               <Out> ByRef value As IntPtr) As Integer

    Private Delegate Function IOcrWord_GetBoundingRect(<[In]> this As IntPtr,
                                                       <Out> ByRef value As RectangleF) As Integer

    Private Delegate Function IOcrWord_GetText(<[In]> this As IntPtr,
                                               <Out> ByRef value As IntPtr) As Integer

    Private Delegate Function IBitmapDecoderStatics_CreateAsync(<[In]> this As IntPtr,
                                                                <[In]> stream As IntPtr,
                                                                <Out> ByRef asyncInfo As IntPtr) As Integer

    Private Delegate Function IBitmapFrame_GetPixelWidth(<[In]> this As IntPtr,
                                                         <Out> ByRef value As Integer) As Integer

    Private Delegate Function IBitmapFrame_GetPixelHeight(<[In]> this As IntPtr,
                                                          <Out> ByRef value As Integer) As Integer

    Private Delegate Function IBitmapFrameWithSoftwareBitmap_GetSoftwareBitmapAsync(<[In]> this As IntPtr,
                                                                                    <Out> ByRef value As IntPtr) As Integer

    Private Delegate Function IAsyncInfo_GetStatus(<[In]> this As IntPtr,
                                                    <Out> ByRef status As AsyncStatus) As Integer

    Private Delegate Function IAsyncInfo_Close(<[In]> this As IntPtr) As Integer

    Private Delegate Function IAsyncInterface_GetResults(<[In]> this As IntPtr,
                                                         <Out> ByRef results As IntPtr) As Integer

    Private Delegate Function IClosable_Close(<[In]> this As IntPtr) As Integer
#End Region

#Region "Interface Functions"
    Private Function ReadOnlyList_GetAt(this As IntPtr,
                                        index As Integer,
                                        ByRef value As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IReadOnlyList_GetAt, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IReadOnlyList_GetAt)),
                                                           IReadOnlyList_GetAt).Invoke(this, index, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function ReadOnlyList_GetSize(this As IntPtr,
                                          ByRef value As Integer) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IReadOnlyList_GetSize, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IReadOnlyList_GetSize)),
                                                           IReadOnlyList_GetSize).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function Language_GetLanguageTag(this As IntPtr,
                                             ByRef value As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.ILanguage_GetLanguageTag, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(ILanguage_GetLanguageTag)),
                                                           ILanguage_GetLanguageTag).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function Language_GetDisplayName(this As IntPtr,
                                             ByRef value As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.ILanguage_GetDisplayName, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(ILanguage_GetDisplayName)),
                                                           ILanguage_GetDisplayName).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function Language_GetNativeName(this As IntPtr,
                                            ByRef value As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.ILanguage_GetNativeName, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(ILanguage_GetNativeName)),
                                                           ILanguage_GetNativeName).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function Language_GetScript(this As IntPtr,
                                        ByRef value As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.ILanguage_GetScript, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(ILanguage_GetScript)),
                                                           ILanguage_GetScript).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function OcrEngineStatics_GetMaxImageDimension(this As IntPtr,
                                                           ByRef value As Integer) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IOcrEngineStatics_GetMaxImageDimension, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IOcrEngineStatics_GetMaxImageDimension)),
                                                           IOcrEngineStatics_GetMaxImageDimension).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function OcrEngineStatics_GetAvailableRecognizerLanguages(this As IntPtr,
                                                                      ByRef value As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IOcrEngineStatics_GetAvailableRecognizerLanguages, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IOcrEngineStatics_GetAvailableRecognizerLanguages)),
                                                           IOcrEngineStatics_GetAvailableRecognizerLanguages).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function OcrEngineStatics_IsLanguageSupported(this As IntPtr,
                                                          language As IntPtr,
                                                          ByRef value As Boolean) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IOcrEngineStatics_IsLanguageSupported, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IOcrEngineStatics_IsLanguageSupported)),
                                                           IOcrEngineStatics_IsLanguageSupported).Invoke(this, language, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function OcrEngineStatics_TryCreateFromLanguage(this As IntPtr,
                                                            language As IntPtr,
                                                            ByRef result As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IOcrEngineStatics_TryCreateFromLanguage, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IOcrEngineStatics_TryCreateFromLanguage)),
                                                           IOcrEngineStatics_TryCreateFromLanguage).Invoke(this, language, result) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function OcrEngine_RecognizeAsync(this As IntPtr,
                                              bitmap As IntPtr,
                                              ByRef result As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IOcrEngine_RecognizeAsync, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IOcrEngine_RecognizeAsync)),
                                                           IOcrEngine_RecognizeAsync).Invoke(this, bitmap, result) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function OcrEngine_GetRecognizerLanguage(this As IntPtr,
                                                     ByRef value As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IOcrEngine_GetRecognizerLanguage, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IOcrEngine_GetRecognizerLanguage)),
                                                           IOcrEngine_GetRecognizerLanguage).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function OcrResult_GetLines(this As IntPtr,
                                        ByRef value As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IOcrResult_GetLines, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IOcrResult_GetLines)),
                                                           IOcrResult_GetLines).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function OcrResult_GetTextAngle(this As IntPtr,
                                            ByRef value As Double) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IOcrResult_GetTextAngle, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IOcrResult_GetTextAngle)),
                                                           IOcrResult_GetTextAngle).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function OcrResult_GetText(this As IntPtr,
                                       ByRef value As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IOcrResult_GetText, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IOcrResult_GetText)),
                                                           IOcrResult_GetText).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function OcrLine_GetWords(this As IntPtr,
                                      ByRef value As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IOcrLine_GetWords, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IOcrLine_GetWords)),
                                                           IOcrLine_GetWords).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function OcrLine_GetText(this As IntPtr,
                                     ByRef value As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IOcrLine_GetText, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IOcrLine_GetText)),
                                                           IOcrLine_GetText).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function OcrWord_GetBoundingRect(this As IntPtr,
                                             ByRef value As RectangleF) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IOcrWord_GetBoundingRect, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IOcrWord_GetBoundingRect)),
                                                           IOcrWord_GetBoundingRect).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function OcrWord_GetText(this As IntPtr,
                                     ByRef value As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IOcrWord_GetText, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IOcrWord_GetText)),
                                                           IOcrWord_GetText).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function BitmapDecoderStatics_CreateAsync(this As IntPtr,
                                                      stream As IntPtr,
                                                      ByRef asyncInfo As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IBitmapDecoderStatics_CreateAsync, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IBitmapDecoderStatics_CreateAsync)),
                                                           IBitmapDecoderStatics_CreateAsync).Invoke(this, stream, asyncInfo) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function BitmapFrame_GetPixelWidth(this As IntPtr,
                                               ByRef value As Integer) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IBitmapFrame_GetPixelWidth, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IBitmapFrame_GetPixelWidth)),
                                                           IBitmapFrame_GetPixelWidth).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function BitmapFrame_GetPixelHeight(this As IntPtr,
                                                ByRef value As Integer) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IBitmapFrame_GetPixelHeight, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IBitmapFrame_GetPixelHeight)),
                                                           IBitmapFrame_GetPixelHeight).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function BitmapFrameWithSoftwareBitmap_GetSoftwareBitmapAsync(this As IntPtr,
                                                                          ByRef value As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IBitmapFrameWithSoftwareBitmap_GetSoftwareBitmapAsync, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IBitmapFrameWithSoftwareBitmap_GetSoftwareBitmapAsync)),
                                                           IBitmapFrameWithSoftwareBitmap_GetSoftwareBitmapAsync).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function AsyncInfo_GetStatus(this As IntPtr,
                                         ByRef status As AsyncStatus) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IAsyncInfo_GetStatus, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IAsyncInfo_GetStatus)),
                                                           IAsyncInfo_GetStatus).Invoke(this, status) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function AsyncInfo_Close(this As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IAsyncInfo_Close, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IAsyncInfo_Close)),
                                                           IAsyncInfo_Close).Invoke(this) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function AsyncInterface_GetResults(this As IntPtr,
                                               ByRef results As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IAsyncInterface_GetResults, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IAsyncInterface_GetResults)),
                                                           IAsyncInterface_GetResults).Invoke(this, results) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function Closable_Close(this As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IClosable_Close, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IClosable_Close)),
                                                           IClosable_Close).Invoke(this) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function
#End Region

#Region "Implements IDisposable"
    Protected Overridable Sub Dispose(disposing As Boolean)
        If Not m_bolDisposedValue Then
            If disposing Then
            End If
            Release(m_pIOcrEngine)
            Release(m_pIBitmapDecoderStatics)
            Release(m_pIOcrEngineStatics)
            m_bolDisposedValue = True
        End If
    End Sub

    Public Sub Dispose() Implements IDisposable.Dispose
        Dispose(disposing:=True)
        GC.SuppressFinalize(Me)
    End Sub
#End Region

End Class
